/***************************************************************************
                          arrayitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Aug 28 1999
    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <klocale.h>
#include <kurl.h>
#include <ksimpleconfig.h>
#include <qfileinfo.h>
#include <kmessagebox.h>
#include <kio/netaccess.h>
#include "arrayitem.h"
#include "arraydlg.h"
#include "kpldoc.h"
#include "kplgraph.h"
#include "kplchecklistitem.h"
#include "utils.h"

ArrayItem::ArrayItem()
{
  init();
  ix = iy = ie = errbars = 0;
}

ArrayItem::ArrayItem(Kpl::AutoStruct* aut) :  ScaledItem(aut)
{
  init();
  ix = aut->ixAuto;
  iy = aut->iyAuto;
  ie = aut->ieAuto;
  errbars = aut->autoErr;
}

ArrayItem::ArrayItem(KSimpleConfig* plo, Kpl::AutoStruct* aut,
                     const KURL& url) : ScaledItem(plo, aut)
{
  init();
  strcpy(path, plo->readEntry("path", ""));
  Utils::absPath(url, path);
  if (!readFile()) {
    KMessageBox::error(0, path);
    return;
  }
  int icolmax = ncols - 1;
  ix = plo->readNumEntry("ix", QMIN(aut->ixAuto, icolmax));
  iy = plo->readNumEntry("iy", QMIN(aut->iyAuto, icolmax));
  ie = plo->readNumEntry("ie", QMIN(aut->ieAuto, icolmax));
  istart = plo->readNumEntry("istart");
  n = plo->readNumEntry("n", nrows - istart);
  errbars = plo->readNumEntry("errbars", aut->autoErr);
}

ArrayItem::~ArrayItem()
{
  freeX();
}

Kpl::ItemTypes ArrayItem::iType() const
{
  return Kpl::Array;
}

void ArrayItem::init()
{
  istart = n = nrows = ncols = 0;
  x = 0;
  path[0] = 0;
}

void ArrayItem::freeX()
{
  if (x)
    delete x[0];
  delete x;
  x = 0;
}

int ArrayItem::readFile()
{
  freeX();
  ncols = nrows = 0;
  KURL url = KURL(path);
  QString fn;
  if (url.isLocalFile())
    fn = url.path();
  else
    if (!KIO::NetAccess::download(url, fn))
      return nrows;
  if (FILE* f = fopen(fn, "r")) {
    int i;
    char s[256];
    do
      if ((i = (fgets(s, 256, f) != 0)))
        i = (s[0] == '#') ? 1 : 2;
    while (i == 1);
    if (i) {
      const char delim[] = " ,\t";
      char* s1 = strtok(s, delim);
      while (s1) {
        ncols++;
        s1 = strtok(0, delim);
      }
      if (ncols) {
        QArray<double> x2;
        rewind(f);
        int np = 0;
        int k = 0;
        while (fgets(s, 256, f))
          if (s[0] != '#') {
            nrows++;
            np += ncols;
            x2.resize(np);
            x2[k++] = ((s1 = strtok(s, delim))) ? atof(s1) : 0.0;
            for (i = 1; i < ncols; i++)
              x2[k++] = ((s1 = strtok(0, delim))) ? atof(s1) : 0.0;
          }
        double* x1 = new double[np];
        x = new double*[ncols];
        for (i = 0; i < ncols; i++)
          x[i] = &x1[i * nrows];
        k = 0;
        for (int j = 0; j < nrows; j++)
          for (i = 0; i < ncols; i++)
            x[i][j] = x2[k++];
      }
    }
    fclose(f);
  }
  if (!url.isLocalFile())
    KIO::NetAccess::removeTempFile(fn);
  return nrows;
}

void ArrayItem::draw(KplGraph* g)
{
  if (x && active) {
    setProperties(g);
    g->plArray(&x[ix][istart], &x[iy][istart], fx, fy, n);
    if (errbars)
      g->plError(&x[ix][istart], &x[iy][istart], &x[ie][istart], fx, fy, n);
  }
}

void ArrayItem::writePlo(FILE* f, const KURL& url, bool abs) const
{
  fprintf(f, "Type=ARRAYITEM\n");
  ScaledItem::writePlo(f, url, abs);
  char rel[260];
  fprintf(f, "ix=%i\niy=%i\nie=%i\nistart=%i\nn=%i\nerrbars=%i\npath=%s\n",
          ix, iy, ie, istart, n, errbars, Utils::relPath(url, rel, path, abs));
}

void ArrayItem::setText(KplCheckListItem* it, bool* arrays, bool*) const
{
  *arrays = true;
  it->setText(1, i18n("Array"));
  QFileInfo fi(path);
  it->setText(2, i18n("file") + " " + fi.fileName() + ", " + i18n("columns")
              + " " + QString::number(ix) + ", " + QString::number(iy));
}

int ArrayItem::editItem(QWidget* parent, KplDoc* m, int)
{
  ArrayDlg dlg(parent, m, this);
  return dlg.exec();
}

double** ArrayItem::copyX() const
{
  int i = ncols * nrows;
  double* x1 = new double[i];
  memcpy(x1, x[0], i * sizeof(double));
  double** x2 = new double*[ncols];
  for (i = 0; i < ncols; i++)
    x2[i] = &x1[i * nrows];
  return x2;
}

KplItem* ArrayItem::copy() const
{
  ArrayItem* it = new ArrayItem(*this);
  if (x)
    it->x = copyX();
  return it;
}

void ArrayItem::expoItem(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (x) {
    Utils::expoFile(iext, fxt, &x[ix][istart], n);
    Utils::expoFile(ieyt, fyt, &x[iy][istart], n);
  }
}

void ArrayItem::minMax(double* xmi, double* xma, double* ymi,
                       double* yma) const
{
  if (x) {
    Utils::minMaxFile(xmi, xma, &x[ix][istart], n);
    Utils::minMaxFile(ymi, yma, &x[iy][istart], n);
  }
}
