/***************************************************************************
                          funitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dlfcn.h>
#include <klocale.h>
#include <ksimpleconfig.h>
#include <qfileinfo.h>
#include "funitem.h"
#include "funcdlg.h"
#include "kplgraph.h"
#include "kplchecklistitem.h"
#include "utils.h"

FunItem::FunItem()
{
  init();
}

FunItem::FunItem(Kpl::AutoStruct* aut) : ScaledItem(aut)
{
  init();
}

FunItem::FunItem(KSimpleConfig* plo, Kpl::AutoStruct* aut, const KURL& url) :
 ScaledItem(plo, aut)
{
  fkty = 0;
  hmody = 0;
  tmin = plo->readDoubleNumEntry("xfmin");
  tmax = plo->readDoubleNumEntry("xfmax", tmin + 1.0);
  dt = plo->readDoubleNumEntry("dx");
  char s1[256];
  strcpy(s1, plo->readEntry("p", "0"));
  QString s2;
  py[0] = ((s2 = strtok(s1, " "))) ? s2.toDouble() : 0.0;
  for (int j = 1; j < KPL_NPMAX; j++)
    py[j] = ((s2 = strtok(0, " "))) ? s2.toDouble() : 0.0;
  strcpy(namey, plo->readEntry("name", ""));
  strcpy(pathy, plo->readEntry("path", ""));
  Utils::absPath(url, pathy);
  Utils::getFuncAddr(pathy, namey, &hmody, &fkty);
}

FunItem::~FunItem()
{
  freeX();
}

Kpl::ItemTypes FunItem::iType() const
{
  return Kpl::Function;
}

void FunItem::init()
{
  tmin = tmax = dt = 0.0;
  fkty = 0;
  memset(py, 0, sizeof(py));
  namey[0] = pathy[0] = 0;
  hmody = 0;
}

void FunItem::freeX()
{
  if (hmody)
    dlclose(hmody);
  hmody = 0;
}

void FunItem::draw(KplGraph* g)
{
  if (fkty && active) {
    setProperties(g);
    bool newv = tv.isNull();
    double d = dt;
    int n;
    if (d)
      n = int(fabs((tmax - tmin) / d)) + 1;
    else {
      d = 0.01 * (tmax - tmin);
      n = 101;
    }
    int i;
    if (newv || (tmin != tmino) || (tmax != tmaxo) || (d != dto)) {
      tv.detach();
      tv.resize(n);
      for (i = 0; i < n; i++)
        tv[i] = tmin + i * d;
    }
    bool pchanged = false;
    for (i = 0; i < KPL_NPMAX; i++)
      if ((pchanged = (py[i] != pyo[i])))
        break;
    if (pchanged || newv || (tmin != tmino) || (tmax != tmaxo) ||
        (d != dto) || (fkty != fktyo)) {
      yv.detach();
      yv.resize(n);
    }
    if (pchanged || newv || (tmin != tmino) || (tmax > tmaxo) ||
        (d != dto) || (fkty != fktyo))
      for (i = 0; i < n; i++)
        yv[i] = ((*fkty)(tv[i], py));
    if (pchanged)
      for (i = 0; i < KPL_NPMAX; i++)
        pyo[i] = py[i];
    if (fkty != fktyo)
      fktyo = fkty;
    if (tmin != tmino)
      tmino = tmin;
    if (tmax != tmaxo)
      tmaxo = tmax;
    if (d != dto)
      dto = d;
    g->plArray(tv.data(), yv.data(), fx, fy, n);
  }
}

void FunItem::writePlo(FILE* f, const KURL& url, bool abs) const
{
  fprintf(f, "Type=FUNITEM\n");
  ScaledItem::writePlo(f, url, abs);
  fprintf(f, "xfmin=%g\nxfmax=%g\ndx=%g\np=", tmin, tmax, dt);
  for (int i = 0; i < 19; i++)
    fprintf(f, "%g ", py[i]);
  char rel[260];
  fprintf(f, "%g\nname=%s\npath=%s\n", py[19], namey,
          Utils::relPath(url, rel, pathy, abs));
}

void FunItem::setText(KplCheckListItem* it, bool*, bool* funcs) const
{
  *funcs = true;
  it->setText(1, i18n("Function"));
  QFileInfo fi(pathy);
  it->setText(2, QString("y = ") + fi.baseName() + "." + namey + "(x)");
}

int FunItem::editItem(QWidget* parent, KplDoc* m, int)
{
  FuncDlg dlg(parent, m, this);
  return dlg.exec();
}

void FunItem::copyX(FunItem* it) const
{
  if (hmody)
    it->hmody = dlopen(it->pathy, RTLD_NOW);
}

KplItem* FunItem::copy() const
{
  FunItem* it = new FunItem(*this);
  copyX(it);
  return it;
}

void FunItem::expoItem(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (fkty) {
    Utils::expo(QMAX(fabs(tmin), fabs(tmax)), iext, fxt);
    Utils::expoFunc(fkty, ieyt, fyt, tmin, tmax, dt, py);
  }
}

void FunItem::minMax(double* xmi, double* xma, double* ymi, double* yma) const
{
  if (fkty) {
    *xmi = tmin;
    *xma = tmax;
    int mtic, ndig;
    double ytic;
    Utils::autoFunc(ymi, yma, &ytic, &mtic, &ndig, fkty, *xmi, *xma, dt, py, 1.0, 0.0);
  }
}

ParFunItem::ParFunItem()
{
  init();
}

ParFunItem::ParFunItem(Kpl::AutoStruct* aut) : FunItem(aut)
{
  init();
}

ParFunItem::ParFunItem(KSimpleConfig* plo, Kpl::AutoStruct* aut,
                       const KURL& url)
{
  fktx = 0;
  hmodx = 0;
  active = plo->readBoolEntry("active", true);
  ScaledItem::readPlo(plo, aut);
  tmin = plo->readDoubleNumEntry("tmin");
  tmax = plo->readDoubleNumEntry("tmax", tmin + 1.0);
  dt = plo->readDoubleNumEntry("dt");
  char s1[256];
  strcpy(s1, plo->readEntry("px", "0"));
  QString s2;
  px[0] = ((s2 = strtok(s1, " "))) ? s2.toDouble() : 0.0;
  for (int j = 1; j < KPL_NPMAX; j++)
    px[j] = ((s2 = strtok(0, " "))) ? s2.toDouble() : 0.0;
  strcpy(s1, plo->readEntry("py", "0"));
  py[0] = ((s2 = strtok(s1, " "))) ? s2.toDouble() : 0.0;
  for (int j = 1; j < KPL_NPMAX; j++)
    py[j] = ((s2 = strtok(0, " "))) ? s2.toDouble() : 0.0;
  strcpy(namex, plo->readEntry("namex", ""));
  strcpy(namey, plo->readEntry("namey", ""));
  strcpy(pathx, plo->readEntry("pathx", ""));
  strcpy(pathy, plo->readEntry("pathy", ""));
  Utils::absPath(url, pathx);
  Utils::absPath(url, pathy);
  Utils::getFuncAddr(pathx, namex, &hmodx, &fktx);
  Utils::getFuncAddr(pathy, namey, &hmody, &fkty);
}

ParFunItem::~ParFunItem()
{
  freeX();
}

Kpl::ItemTypes ParFunItem::iType() const
{
  return Kpl::ParFunction;
}

void ParFunItem::init()
{
  fktx = 0;
  memset(px, 0, sizeof(px));
  namex[0] = pathx[0] = 0;
  hmodx = 0;
}

void ParFunItem::freeX()
{
  if (hmodx)
    dlclose(hmodx);
  if (hmody)
    dlclose(hmody);
  hmodx = hmody = 0;
}

void ParFunItem::draw(KplGraph* g)
{
  if (fkty && active) {
    setProperties(g);
    bool newv = tv.isNull();
    double d = dt;
    int n;
    if (d)
      n = int(fabs((tmax - tmin) / d)) + 1;
    else {
      d = 0.01 * (tmax - tmin);
      n = 101;
    }
    int i;
    if (newv || (tmin != tmino) || (tmax != tmaxo) || (d != dto)) {
      tv.detach();
      tv.resize(n);
      for (i = 0; i < n; i++)
        tv[i] = tmin + i * d;
    }
    bool pchanged = false;
    for (i = 0; i < KPL_NPMAX; i++)
      if ((pchanged = (py[i] != pyo[i])))
        break;
    if (pchanged || newv || (tmin != tmino) || (tmax != tmaxo) ||
        (d != dto) || (fkty != fktyo)) {
      yv.detach();
      yv.resize(n);
    }
    if (pchanged || newv || (tmin != tmino) || (tmax > tmaxo) ||
        (d != dto) || (fkty != fktyo))
      for (i = 0; i < n; i++)
        yv[i] = ((*fkty)(tv[i], py));
    if (pchanged) {
      for (i = 0; i < KPL_NPMAX; i++)
        pyo[i] = py[i];
      pchanged = false;
    }
    for (i = 0; i < KPL_NPMAX; i++)
      if ((pchanged = (px[i] != pxo[i])))
        break;
    if (pchanged || newv || (tmin != tmino) || (tmax != tmaxo) ||
        (d != dto) || (fktx != fktxo)) {
      xv.detach();
      xv.resize(n);
    }
    if (pchanged || newv || (tmin != tmino) || (tmax > tmaxo) ||
        (d != dto) || (fktx != fktxo))
      for (i = 0; i < n; i++)
        xv[i] = ((*fktx)(tv[i], px));
    if (pchanged)
      for (i = 0; i < KPL_NPMAX; i++)
        pxo[i] = px[i];
    if (fktx != fktxo)
      fktxo = fktx;
    if (fkty != fktyo)
      fktyo = fkty;
    if (tmin != tmino)
      tmino = tmin;
    if (tmax != tmaxo)
      tmaxo = tmax;
    if (d != dto)
      dto = d;
    g->plArray(xv.data(), yv.data(), fx, fy, n);
  }
}

void ParFunItem::writePlo(FILE* f, const KURL& url, bool abs) const
{
  fprintf(f, "Type=PARFUNITEM\n");
  ScaledItem::writePlo(f, url, abs);
  fprintf(f, "tmin=%g\ntmax=%g\ndt=%g\npx=", tmin, tmax, dt);
  int i;
  for (i = 0; i < 19; i++)
    fprintf(f, "%g ", px[i]);
  fprintf(f, "%g\npy=", px[19]);
  for (i = 0; i < 19; i++)
    fprintf(f, "%g ", py[i]);
  fprintf(f, "%g\nnamex=%s\nnamey=%s\n", py[19], namex, namey);
  char rel[260];
  fprintf(f, "pathx=%s", Utils::relPath(url, rel, pathx, abs));
  fprintf(f, "pathy=%s", Utils::relPath(url, rel, pathy, abs));
}

void ParFunItem::setText(KplCheckListItem* it, bool*, bool*) const
{
  it->setText(1, i18n("Par. function"));
  QFileInfo fi1(pathx);
  QFileInfo fi2(pathy);
  it->setText(2, QString("x = ") + fi1.baseName() + "." + namex
              + "(t), y = " + fi2.baseName() + "." + namey + "(t)");
}

int ParFunItem::editItem(QWidget* parent, KplDoc* m, int)
{
  ParFuncDlg dlg(parent, m, this);
  return dlg.exec();
}

void ParFunItem::copyX(ParFunItem* it) const
{
  if (hmodx)
    it->hmodx = dlopen(it->pathx, RTLD_NOW);
  if (hmody)
    it->hmody = dlopen(it->pathy, RTLD_NOW);
}

KplItem* ParFunItem::copy() const
{
  ParFunItem* it = new ParFunItem(*this);
  copyX(it);
  return it;
}

void ParFunItem::expoItem(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (fktx && fkty) {
    Utils::expoFunc(fktx, iext, fxt, tmin, tmax, dt, px);
    Utils::expoFunc(fkty, ieyt, fyt, tmin, tmax, dt, py);
  }
}

void ParFunItem::minMax(double* xmi, double* xma, double* ymi,
                        double* yma) const
{
  if (fktx && fkty) {
    int mtic, ndig;
    double tic;
    Utils::autoFunc(xmi, xma, &tic, &mtic, &ndig, fktx, tmin, tmax, dt, px, 1.0, 0.0);
    Utils::autoFunc(ymi, yma, &tic, &mtic, &ndig, fkty, tmin, tmax, dt, py, 1.0, 0.0);
  }
}
