/***************************************************************************
                          kpl.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <qprinter.h>
#include <qscrollbar.h>
#include <qimage.h>
#include <qdir.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kfiledialog.h>
#include <kprocess.h>
#include <kkeydialog.h>
#include <kmenubar.h>
#include <kconfig.h>
#include <kapp.h>
#include <kmessagebox.h>
#include <kio/job.h>
#include <kstddirs.h>
#include <kaction.h>
#include <kstdaction.h>
#include <khelpmenu.h>
#include <kedittoolbar.h>
#include "kplspinbox.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplview.h"
#include "kplwidget.h"
#include "zoomdlg.h"
#include "refreshdlg.h"
#include "basicdlg.h"
#include "autodlg.h"
#include "itemdlg.h"
#include "autofitdlg.h"
#include "psgraph.h"
#include "kplitem.h"
#include "utils.h"

KplApp::KplApp() : KMainWindow(0)
{
  config = kapp->config();
  plot = false;
  itemDlg = 0;
  tRefresh = 5;
  setMinimumSize(120, 80);
  initDocument();
  initActions();
  readOptions();
  initView();
  initKeyAccel();
  calcStatusBar();
  initStatusBar();
  updWin();
  timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), SLOT(slotViewReload()));
  connect(kapp, SIGNAL(lastWindowClosed()), kapp, SLOT(quit()));
  connect(doc, SIGNAL(modelChanged(bool, bool)), SLOT(slotUpdate(bool, bool)));
  connect(view->page, SIGNAL(mouseMoved(int, int)), SLOT(updStatus(int, int)));
  connect(view->page, SIGNAL(mouseLeaved()), SLOT(clearStatus()));
  connect(view->page, SIGNAL(rightButtonPressed()), SLOT(showPopupMenu()));
  connect(view->page, SIGNAL(urlsDropped(QStrList *)),
          SLOT(slotOpenList(QStrList *)));
}

KplApp::~KplApp()
{
  delete key_accel;
  delete itemDlg;
}

void KplApp::openDocumentURL(const KURL& url)
{
  statusBar()->message(i18n("Opening file..."));
  if (warning->isChecked() ? queryClose() : true)
    doc->openDocument(url);
  slotStatusMsg(i18n("Ready."));
}

KplDoc* KplApp::getDocument() const
{
	return doc;
}

void KplApp::initActions()
{
  actionCollection()->setHighlightingEnabled(true);
  connect(actionCollection(), SIGNAL(actionStatusText(const QString &)),
          this, SLOT(slotStatusHelpMsg(const QString &)));
  KAction* a = KStdAction::openNew(this, SLOT(slotFileNew()),
                                   actionCollection());
  a->setStatusText(i18n("Creates a new plot"));
  a = new KAction(i18n("&Open Plot File..."), QIconSet(BarIcon("fileopen")),
                  KStdAccel::open(), this, SLOT(slotFileOpenPlot()),
                  actionCollection(), "open_plot");
  a->setStatusText(i18n("Opens an existing plot file"));
  a = new KAction(i18n("Open &Data File..."), QIconSet(BarIcon("fileopen")), 0,
                  this, SLOT(slotFileOpenData()), actionCollection(),
                  "open_data");
  a->setStatusText(i18n("Opens an existing data file"));
  recentFiles = KStdAction::openRecent(this,
                                       SLOT(openDocumentURL(const KURL&)),
                                       actionCollection());
  fileSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
  fileSave->setStatusText(i18n("Saves the current plot"));
  fileSaveAs = KStdAction::saveAs(this, SLOT(slotFileSaveAs()),
                                  actionCollection());
  fileSaveAs->setStatusText(i18n("Saves the plot under a different name"));
  fileClose = KStdAction::close(this, SLOT(slotFileClose()),
                                actionCollection());
  fileClose->setStatusText(i18n("Closes the current plot"));
  filePlot = new KAction(i18n("Display P&lot"), 0, this, SLOT(slotFilePlot()),
                         actionCollection(), "plot");
  filePlot->setStatusText(i18n("Displays the current plot"));
  filePS = new KListAction(i18n("PostScript O&utput"), 0,
                           actionCollection(), "postscript");
  QStringList list;
  list.append(i18n("&Portrait..."));
  list.append(i18n("&Landscape..."));
  filePS->setItems(list);
  connect(filePS, SIGNAL(activated(int)), this, SLOT(slotFilePS(int)));
  filePrint = KStdAction::print(this, SLOT(slotFilePrint()),
                                actionCollection());
  filePrint->setStatusText(i18n("Prints the current plot"));
  a = new KAction(i18n("New &Window"), QIconSet(kapp->miniIcon()), 0, this,
                  SLOT(slotFileNewWindow()), actionCollection(), "new_window");
  a->setStatusText(i18n("Opens a new application window"));
  a = new KAction(i18n("Close W&indow"), 0, this, SLOT(slotFileCloseWindow()),
                  actionCollection(), "close_window");
  a->setStatusText(i18n("Closes the current window"));
  a = KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());
  a->setStatusText(i18n("Exits the program"));
  editUndo = KStdAction::undo(this, SLOT(slotEditUndo()), actionCollection());
  editUndo->setStatusText(i18n("Undoes the last change"));
  editRedo = KStdAction::redo(this, SLOT(slotEditRedo()), actionCollection());
  editRedo->setStatusText(i18n("Redoes the last change"));
  a = new KAction(i18n("&Items..."), 0, this, SLOT(slotEditItems()),
                  actionCollection(), "items");
  a->setStatusText(i18n("Edits plot items"));
  a = KStdAction::zoomIn(this, SLOT(slotViewZoomIn()), actionCollection());
  a->setStatusText(i18n("Increases magnification by 20 %"));
  a = KStdAction::zoomOut(this, SLOT(slotViewZoomOut()), actionCollection());
  a->setStatusText(i18n("Decreases magnification by 20 %"));
  a = KStdAction::zoom(this, SLOT(slotViewZoom()), actionCollection());
  a->setStatusText(i18n("Sets magnification"));
  viewReload = KStdAction::redisplay(this, SLOT(slotViewReload()),
                                     actionCollection());
  viewReload->setAccel(KStdAccel::reload());
  viewReload->setStatusText(i18n("Reloads file"));
  a = new KAction(i18n("Reload &Periodically..."), 0, this,
                  SLOT(slotViewAutoReload()), actionCollection(),
                  "autoreload");
  a->setStatusText(i18n("Reloads file periodically"));
  showToolbar = KStdAction::showToolbar(this, SLOT(slotOptionsToolBar()),
                                        actionCollection());
  showToolbar->setStatusText(i18n("Enables / disables the toolbar"));
  showStatusbar = KStdAction::showStatusbar(this, SLOT(slotOptionsStatusBar()),
                                            actionCollection());
  showStatusbar->setStatusText(i18n("Enables / disables the statusbar"));
  autoplot = new KToggleAction(i18n("Aut&oplot"), 0, this,
                               SLOT(slotOptionsAutoPlot()),
                               actionCollection(), "autoplot");
  autoplot->setStatusText(i18n("Enables / disables automatic plotting"));
  showSource = new KToggleAction(i18n("S&how Function Source"), 0, this,
                                 SLOT(slotOptionsShowSource()),
                                 actionCollection(), "show_source");
  showSource->setStatusText(i18n("Enables / disables display of "
                                 "function source"));
  boundingBox = new KToggleAction(i18n("Calculate &PS Bounding Box"), 0,
                                 actionCollection(), "bounding_box");
  boundingBox->setStatusText(i18n("Enables / disables calculation of "
                                  "PS bounding box"));
  saveAbsPath = new KToggleAction(i18n("Save Absol&ute Paths"), 0,
                              actionCollection(), "save_abs_path");
  saveAbsPath->setStatusText(i18n("Saves absolute paths of data "
                                  "and library files in plot files"));
  warning = new KToggleAction(i18n("Unsaved Changes &Warning"), 0,
                              actionCollection(), "warning");
  warning->setStatusText(i18n("Enables / disables warning on "
                              "unsaved changes"));
  saveAtEnd = new KToggleAction(i18n("Save Options At &End"), 0,
                                actionCollection(), "save_at_end");
  saveAtEnd->setStatusText(i18n("Enables / disables saving options at end"));
  a = KStdAction::keyBindings(this, SLOT(slotOptionsKeys()),
                              actionCollection());
  a->setStatusText(i18n("Displays key bindings dialog"));
  a = KStdAction::configureToolbars(this, SLOT(slotEditToolbars()),
                                    actionCollection());
  a->setStatusText(i18n("Displays toolbar configuration dialog"));
  a = new KAction(i18n("Configure Bas&ics..."), 0,
                  this, SLOT(slotOptionsBasic()),
                  actionCollection(), "configure_basic");
  a->setStatusText(i18n("Displays basic settings dialog"));
  a = new KAction(i18n("&Configure Autoplot..."), 0, this,
                  SLOT(slotOptionsAuto()),
                  actionCollection(), "configure_auto");
  a->setStatusText(i18n("Displays autoplot settings dialog"));
  a = new KAction(i18n("Configure Auto&fit..."), 0, this,
                  SLOT(slotOptionsAutoFit()),
                  actionCollection(), "configure_autofit");
  a->setStatusText(i18n("Displays autofit settings dialog"));
  a = KStdAction::saveOptions(this, SLOT(slotOptionsSave()),
                              actionCollection());
  a->setStatusText(i18n("Saves options"));
  new KHelpMenu(this, instance()->aboutData(), false, actionCollection());
}

void KplApp::calcStatusBar()
{
  xPrec = QMIN(QMAX(2, Utils::int9(log10(QMAX(fabs(view->xmin),
                                              fabs(view->xmax)) / view->fxn
                                   + 1.0e-37) + 0.999999) +
               Utils::int9(log10(view->xscal * view->fxn + 1.0e-37) +
                           0.999999)), 16);
  yPrec = QMIN(QMAX(2, Utils::int9(log10(QMAX(fabs(view->ymin),
                                              fabs(view->ymax)) / view->fyn
                                   + 1.0e-37) + 0.999999) +
               Utils::int9(log10(-view->yscal * view->fyn + 1.0e-37) +
                           0.999999)), 16);
}

void KplApp::initStatusBar()
{
  statusBar()->insertFixedItem(" 10.00 cm,  15.00 cm ", StatusPos, true);
  statusBar()->insertFixedItem("  x = " +
                               QString::number(-10.0E-33 / 3.0, 'g', xPrec)
                               + ",  y = " +
                               QString::number(-10.0E-33 / 3.0, 'g', yPrec),
                               StatusVal, true);
  statusBar()->insertItem(i18n("Ready."), StatusMsg);
  statusBar()->changeItem("", StatusVal);
  statusBar()->changeItem("", StatusPos);
}

void KplApp::newStatusBar()
{
  int xold = xPrec;
  int yold = yPrec;
  calcStatusBar();
  if ((xPrec != xold) || (yPrec != yold)) {
    statusBar()->changeItem("  x = " +
                            QString::number(-10.0E-33 / 3.0, 'g', xPrec) +
                            ",  y = " +
                            QString::number(-10.0E-33 / 3.0, 'g', yPrec),
                            StatusVal);
    statusBar()->setItemFixed(StatusVal);
    statusBar()->changeItem("", StatusVal);
  }
}

void KplApp::initKeyAccel()
{
  key_accel = new KAccel(this);
  KAction* a = new KAction(i18n("Scroll Left"), Key_Left, view,
                           SLOT(slotScrollLeft()), actionCollection(),
                           "cursor_left");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Right"), Key_Right, view,
                  SLOT(slotScrollRight()), actionCollection(), "cursor_right");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Up"), Key_Up, view, SLOT(slotScrollUp()),
                  actionCollection(), "cursor_up");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Down"), Key_Down, view, SLOT(slotScrollDown()),
               actionCollection(), "cursor_down");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Page Up"), Key_PageUp, view,
                  SLOT(slotScrollPageUp()), actionCollection(), "page_up");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Page Down"), Key_PageDown, view,
                  SLOT(slotScrollPageDown()), actionCollection(), "page_down");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Page Left"), CTRL + Key_PageUp, view,
                  SLOT(slotScrollPageLeft()), actionCollection(), "page_left");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll Page Right"), CTRL + Key_PageDown, view,
                  SLOT(slotScrollPageRight()), actionCollection(),
                  "page_right");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll to Left Margin"), CTRL + Key_Home, view,
                  SLOT(slotScrollHomeLeft()), actionCollection(),
                  "margin_left");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll to Bottom Margin"), Key_Home, view,
                  SLOT(slotScrollHomeDown()), actionCollection(),
                  "margin_bottom");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll to Right Margin"), CTRL + Key_End, view,
                  SLOT(slotScrollEndRight()), actionCollection(),
                  "margin_right");
  a->plugAccel(key_accel);
  a = new KAction(i18n("Scroll to Top Margin"), Key_End, view,
                  SLOT(slotScrollEndUp()), actionCollection(), "margin_top");
  a->plugAccel(key_accel);
  createGUI();
  if (!showToolbar->isChecked())
    toolBar("mainToolBar")->hide();
}

void KplApp::initDocument()
{
  doc = new KplDoc(this);
  doc->newDocument();
}

void KplApp::initView()
{
  view = new KplView(this, fz);
  setCentralWidget(view);
}

void KplApp::saveOptions(bool saveAll)
{
  recentFiles->saveEntries(config);
  if (itemDlg) {
    config->setGroup("ItemDialog");
    config->writeEntry("x", itemDlg->x());
    config->writeEntry("y", itemDlg->y());
    config->writeEntry("Width", itemDlg->width());
    config->writeEntry("Height", itemDlg->height());
  }
  config->setGroup(kapp->instanceName());
	if (saveAll) {
    config->writeEntry("ShowToolbar", showToolbar->isChecked());
    config->writeEntry("ShowStatusbar", showStatusbar->isChecked());
    config->writeEntry("ShowItemDialog", itemDlg ?
                       itemDlg->isVisible() : false);
    config->writeEntry("Directory", doc->currentDir());
    config->writeEntry("Width", width());
    config->writeEntry("Height", height());
    config->writeEntry("AutoPlot", autoplot->isChecked());
    config->writeEntry("ShowSource", showSource->isChecked());
    config->writeEntry("BoundingBox", boundingBox->isChecked());
    config->writeEntry("SaveAbsPath", saveAbsPath->isChecked());
    config->writeEntry("Warning", warning->isChecked());
    config->writeEntry("SaveAtEnd", saveAtEnd->isChecked());
    config->writeEntry("AutoIndexX", doc->aut.ixAuto);
    config->writeEntry("AutoIndexY", doc->aut.iyAuto);
    config->writeEntry("AutoIndexE", doc->aut.ieAuto);
    config->writeEntry("AutoNorm", doc->aut.autoNorm);
    config->writeEntry("AutoErrBar", doc->aut.autoErr);
    config->writeEntry("AutoGrid", doc->aut.autoGrid);
    config->writeEntry("AutoPathHeader", doc->aut.autohPath);
    config->writeEntry("AutoFit", doc->aut.autoFit);
    config->writeEntry("AutoSymbol", doc->aut.autoSymb);
    config->writeEntry("AutoX0", doc->aut.x0Auto);
    config->writeEntry("AutoY0", doc->aut.y0Auto);
    config->writeEntry("AutoXl", doc->aut.xlAuto);
    config->writeEntry("AutoYl", doc->aut.ylAuto);
    config->writeEntry("FitErrorColumn", doc->aut.err.fitErrCol);
    config->writeEntry("FitNonLinear", doc->aut.fitNonLin);
    config->writeEntry("FitShowDialog", doc->aut.fitShowDlg);
    config->writeEntry("FitSaveParam", doc->aut.fitSavePar);
    config->writeEntry("FitFollow", doc->aut.fitFollow);
    config->writeEntry("FitMaxIterations", doc->aut.fitMaxIt);
    config->writeEntry("FitTolerance", doc->aut.fitTol);
    config->writeEntry("FitSymbol", doc->aut.fitSymb);
    config->writeEntry("FitColor", Utils::rgbQt1(doc->aut.fitColor));
    config->writeEntry("FitPath", doc->aut.fitPath);
    config->writeEntry("FitName", doc->aut.fitName);
    config->writeEntry("ErrorModelPath", doc->aut.err.errModPath);
    config->writeEntry("ErrorModelName", doc->aut.err.errModName);
    QStrList s1, s2, s3;
    for (int i = 0; i < KPL_NPMAX; i++) {
      s1.insert(i, QString::number((int) doc->aut.fitPar[i]));
      s2.insert(i, QString::number(doc->aut.pFit[i], doc->aut.format,
                                   doc->aut.prec));
      s3.insert(i, QString::number(doc->aut.err.pErrMod[i], doc->aut.format,
                                   doc->aut.prec));
    }
    config->writeEntry("FitEnableParam", s1, ' ');
    config->writeEntry("FitParam", s2, ' ');
    config->writeEntry("ErrorModelParam", s3, ' ');
    config->writeEntry("FormatX", doc->aut.xf);
    config->writeEntry("FormatY", doc->aut.yf);
    config->writeEntry("ColorFrame", Utils::rgbQt1(doc->aut.colFrame));
    config->writeEntry("ColorGrid", Utils::rgbQt1(doc->aut.colGrid));
    config->writeEntry("ColorData", Utils::rgbQt1(doc->aut.colData));
    config->writeEntry("OutputFormat", QString(QChar(doc->aut.format)));
    config->writeEntry("Precision", doc->aut.prec);
    config->writeEntry("Zoom", view->fZoom);
    config->writeEntry("AutoLetX", doc->aut.sxAuto);
    config->writeEntry("AutoLetY", doc->aut.syAuto);
    config->writeEntry("AutoLetH", doc->aut.shAuto);
  }
  config->sync();
}

void KplApp::readOptions()
{
  config->setGroup(kapp->instanceName());
  config->setDollarExpansion();
  mainWidth = config->readNumEntry("Width", 640);
  mainHeight = config->readNumEntry("Height", 512);
  resize(mainWidth, mainHeight);
  recentFiles->loadEntries(config);
  showToolbar->setChecked(config->readBoolEntry("ShowToolbar", true));
  showStatusbar->setChecked(config->readBoolEntry("ShowStatusbar", true));
  if (!showStatusbar->isChecked())
    statusBar()->hide();
  if (config->readBoolEntry("ShowItemDialog", false)) {
    slotEditItems();
    config->setGroup(kapp->instanceName());
  }
  QString kplDir = KGlobal::dirs()->findResourceDir("data",
    kapp->instanceName() + "/demo1.plo") + QString(kapp->instanceName());
  doc->setCurrentDir(config->readEntry("Directory", kplDir) + "/x");
  doc->aut.showSource = config->readBoolEntry("ShowSource", false);
  showSource->setChecked(doc->aut.showSource);
  autoplot->setChecked(config->readBoolEntry("AutoPlot", true));
  boundingBox->setChecked(config->readBoolEntry("BoundingBox"));
  saveAbsPath->setChecked(config->readBoolEntry("SaveAbsPath", true));
  warning->setChecked(config->readBoolEntry("Warning", true));
  saveAtEnd->setChecked(config->readBoolEntry("SaveAtEnd"));
  doc->aut.ixAuto = config->readNumEntry("AutoIndexX");
  doc->aut.iyAuto = config->readNumEntry("AutoIndexY", 1);
  doc->aut.ieAuto = config->readNumEntry("AutoIndexE", 2);
  doc->aut.autoNorm = config->readBoolEntry("AutoNorm", true);
  doc->aut.autoErr = config->readBoolEntry("AutoErrBar");
  doc->aut.autoGrid = config->readNumEntry("AutoGrid", 2);
  doc->aut.autohPath = config->readBoolEntry("AutoPathHeader", true);
  doc->aut.autoFit = config->readBoolEntry("AutoFit");
  doc->aut.autoSymb = config->readNumEntry("AutoSymbol", 1);
  doc->aut.x0Auto = config->readDoubleNumEntry("AutoX0", 4.0);
  doc->aut.y0Auto = config->readDoubleNumEntry("AutoY0", 3.0);
  doc->aut.xlAuto = config->readDoubleNumEntry("AutoXl", 15.0);
  doc->aut.ylAuto = config->readDoubleNumEntry("AutoYl", 10.0);
  doc->aut.err.fitErrCol = config->readBoolEntry("FitErrorColumn");
  doc->aut.fitNonLin = config->readBoolEntry("FitNonLinear", true);
  doc->aut.fitShowDlg = config->readBoolEntry("FitShowDialog");
  doc->aut.fitSavePar = config->readBoolEntry("FitSaveParam");
  doc->aut.fitFollow = config->readBoolEntry("FitFollow");
  doc->aut.fitMaxIt = config->readNumEntry("FitMaxIterations", 100);
  doc->aut.fitTol = config->readDoubleNumEntry("FitTolerance", 1e-14);
  doc->aut.fitSymb = config->readNumEntry("FitSymbol", 1);
  doc->aut.fitColor = Utils::rgbQt(config->readUnsignedNumEntry("FitColor"));
  strcpy(doc->aut.fitPath, config->readEntry("FitPath", kplDir + "/fkt.so"));
  strcpy(doc->aut.fitName, config->readEntry("FitName", "polynom"));
  strcpy(doc->aut.err.errModPath, config->readEntry("ErrorModelPath",
                                                    kplDir + "/fkt.so"));
  strcpy(doc->aut.err.errModName, config->readEntry("ErrorModelName",
                                                    "polynom"));
  char s1[512];
  strcpy(s1, config->readEntry("FitEnableParam", "0"));
  QCString s2;
  doc->aut.fitPar[0] = ((s2 = strtok(s1, " "))) ? s2.toInt() : false;
  int i;
  for (i = 1; i < KPL_NPMAX; i++)
    doc->aut.fitPar[i] = ((s2 = strtok(0, " "))) ? s2.toInt() : false;
  strcpy(s1, config->readEntry("FitParam", "0"));
  doc->aut.pFit[0] = ((s2 = strtok(s1, " "))) ? s2.toDouble() : 0.0;
  for (i = 1; i < KPL_NPMAX; i++)
    doc->aut.pFit[i] = ((s2 = strtok(0, " "))) ? s2.toDouble() : 0.0;
  strcpy(s1, config->readEntry("ErrorModelParam", "1.e-14 0.001"));
  doc->aut.err.pErrMod[0] = ((s2 = strtok(s1, " "))) ? s2.toDouble() : 0.0;
  for (i = 1; i < KPL_NPMAX; i++)
    doc->aut.err.pErrMod[i] = ((s2 = strtok(0, " "))) ? s2.toDouble() : 0.0;
  doc->aut.xf = config->readDoubleNumEntry("FormatX", 21.0);
  doc->aut.yf = config->readDoubleNumEntry("FormatY", 29.7);
  doc->aut.colFrame = Utils::rgbQt(config->readUnsignedNumEntry("ColorFrame"));
  doc->aut.colGrid = Utils::rgbQt(config->readUnsignedNumEntry("ColorGrid"));
  doc->aut.colData = Utils::rgbQt(config->readUnsignedNumEntry("ColorData"));
  doc->aut.format = *(config->readEntry("OutputFormat", "g").latin1());
  doc->aut.prec = config->readNumEntry("Precision", 6);
  fz = config->readDoubleNumEntry("Zoom", 1.0);
  config->setDollarExpansion(false);
  strcpy(doc->aut.sxAuto, config->readEntry("AutoLetX", "x"));
  strcpy(doc->aut.syAuto, config->readEntry("AutoLetY", "y"));
  strcpy(doc->aut.shAuto, config->readEntry("AutoLetH", ""));
 }

void KplApp::openRead(Kpl::FileTypes type)
{
  if (warning->isChecked())
    if (!queryClose())
      return;
  KURL url = KFileDialog::getOpenURL(doc->currentDir(), type == Kpl::Data ?
                                     "*.dat *.DAT" : "*.plo *.PLO");
  if (!url.isEmpty())
    doc->openDocument(url, type);
}

bool KplApp::save(bool current)
{
  bool success = false;
  if ((doc->currentURL().fileName() != i18n("Untitled")) && current &&
      (doc->fType == Kpl::Plot)) {
    success = doc->saveDocument(doc->currentURL(), saveAbsPath->isChecked());
    if (success)
      updWin();
  } else {
    QString s;
    KURL u;
    if (Utils::getWriteURL(this, u, "*.plo", doc))
      success = doc->saveDocument(u, saveAbsPath->isChecked());
    if (success)
      doc->setCurrentURL(u);
      slotUpdate(false, true);
  }
  return success;
}

bool KplApp::queryClose()
{
  if (doc->isModified())
    do {
      int i = KMessageBox::warningYesNoCancel(this,
        i18n("The current plot has been changed.\nSave file?"));
      if (i == KMessageBox::Cancel)
        return false;
      if (i == KMessageBox::No)
        break;
    } while (!save(true));
  return true;
}

bool KplApp::queryExit()
{
  saveOptions(saveAtEnd->isChecked());
  return true;
}

void KplApp::saveProperties(KConfig* _cfg)
{
  bool untitled = (doc->currentURL().fileName() == i18n("Untitled"));
  if (untitled && !doc->isModified())
    return;
  QString fileName;
  if (untitled)
    fileName = QDir::homeDirPath() + "/" + i18n("Untitled");
  else
    fileName = doc->currentURL().url();
  _cfg->writeEntry("filename", fileName);
  _cfg->writeEntry("modified", doc->isModified());
  if (doc->isModified())
    doc->saveDocument(kapp->tempSaveName(fileName));
}

void KplApp::readProperties(KConfig* _cfg)
{
  QString filename = _cfg->readEntry("filename", "");
  KURL u = KURL(filename);
  bool recovered = false;
  if (_cfg->readBoolEntry("modified")) {
    bool canRecover;
    QString tempname = kapp->checkRecoverFile(filename, canRecover);
    if (canRecover) {
      doc->openDocument(tempname, Kpl::Plot);
      QFile::remove(tempname);
      doc->setModified();
      recovered = true;
    }
  } else
    if (!filename.isEmpty()) {
      doc->openDocument(u);
      recovered = true;
    }
  if (recovered)
    doc->setCurrentURL(u);
  updWin();
}		

void KplApp::slotFilePS(int landscape)
{
  statusBar()->message(i18n("Generating PostScript file..."));
  KURL u;
  if (Utils::getWriteURL(this, u, "*.ps *.eps", doc)) {
    QString filename;
    if (u.isLocalFile())
      filename = u.path();
    else
      filename = QDir::homeDirPath() + "/kpltmp.tmp";
    FILE* f = fopen(filename, "w");
    if (f) {
      view->page->setCursor(waitCursor);
      if (doc->itd.count()) {
        PSGraph g;
        g.psInit(f, landscape ? doc->aut.yf : doc->aut.xf,
                 landscape ? doc->aut.xf : doc->aut.yf);
        if (landscape)
          fprintf(f, "%i 0 translate 90 rotate\n",
                  qRound(doc->aut.yf * Kpl::pts));
        for (KplItem* item = doc->itd.first(); item; item = doc->itd.next())
          item->draw(&g);
        g.psExit();
      }
      if (boundingBox->isChecked()) {
        QCString s2;
        KProcess kp;
        kp << "gs" << "-dNOPAUSE" << "-dBATCH" << "-q" << "-sDEVICE=pbmraw"
           << ("-sOutputFile=" + QDir::homeDirPath() + "/kpltmp.pbm") << "-r72"
           << s2.sprintf("-g%ix%i", qRound(Kpl::pts *
             (landscape ? doc->aut.yf : doc->aut.xf)),
             qRound(Kpl::pts * (landscape ? doc->aut.xf : doc->aut.yf)))
           << filename;
        kp.start(KProcess::Block);
        QImage image;
        if (image.load(QDir::homeDirPath() + "/kpltmp.pbm", "PBMRAW")) {
          int x, y, xl = 0, yo = 0;
          int xr = image.width();
          int yu = image.height();
          for (y = 0; y < image.height(); y++)
            for (x = 0; x < image.width(); x++)
              if (!(image.pixel(x, y) & 0xffffff)) {
                yo = y;
                xl = xr = x;
                goto yo_found;
              }
  yo_found:
          for (y = image.height() - 1; y > yo; y--)
            for (x = 0; x < image.width(); x++)
              if (!(image.pixel(x, y) & 0xffffff)) {
                yu = y;
                xl = QMIN(xl, x);
                xr = QMAX(xr, x);
                goto yu_found;
              }
  yu_found:
          for (y = yo; y <= yu; y++){
            for (x = 0; x < xl; x++)
              if (!(image.pixel(x, y) & 0xffffff)) {
                xl = x;
                break;
              }
            for (x = image.width() - 1; x > xr; x--)
              if (!(image.pixel(x, y) & 0xffffff)) {
                xr = x;
                break;
              }
          }
          xl = QMAX(xl - 2, 0);
          xr = QMIN(xr + 2, image.width() - 1);
          yu = QMAX(image.height() - yu - 2, 0);
          yo = QMIN(image.height() - yo + 2, image.height() - 1);
          remove(QDir::homeDirPath() + "/kpltmp.pbm");
          f = fopen(filename, "r+");
          fprintf(f, "%%!PS-Adobe-3.0 EPSF-3.0\n%%%%BoundingBox: %i %i %i %i      ",
                  xl, yu, xr, yo);
          fclose(f);
        } else
          KMessageBox::error(this,
            i18n("Bitmap file for bounding box calculation not loaded!"));
      }
      doc->setCurrentDir(u);
      if (!u.isLocalFile()) {
        doc->job = KIO::file_copy(KURL(filename), u, -1, true);
        doc->connectIOfinished();
      }
      view->page->setCursor(crossCursor);
    } else
      KMessageBox::sorry(this, i18n("Writing to this file not possible."));
  }
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileNew()
{
  if (queryClose()) {
    doc->newDocument();
    doc->backupItems();
    slotEditItems();
  }
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileOpenPlot()
{
  statusBar()->message(i18n("Opening file..."));
  openRead(Kpl::Plot);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileOpenData()
{
  statusBar()->message(i18n("Opening file..."));
  openRead(Kpl::Data);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileSave()
{
  statusBar()->message(i18n("Saving file..."));
  save(true);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileSaveAs()
{
  statusBar()->message(i18n("Saving file under different name..."));
  save(false);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileClose()
{
  statusBar()->message(i18n("Closing file..."));
  if (queryClose()) {
    doc->newDocument();
    doc->backupItems();
  }
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFilePlot()
{
  statusBar()->message(i18n("Plotting..."));
  plot = doc->itd.count();
  slotPlotCond();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotPlotCond()
{
  if (plot)
    view->paintWidget();
  newStatusBar();
}

void KplApp::slotFilePrint()
{
  statusBar()->message(i18n("Printing..."));
  QPrinter printer;
  if (printer.setup(this))
    view->print(&printer);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileNewWindow()
{
  (new KplApp)->show();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotFileCloseWindow()
{
  close();
}

void KplApp::slotFileQuit()
{
  slotStatusMsg(i18n("Exiting..."));
  if (memberList)
    for (KMainWindow* w = memberList->first(); w; w = memberList->first())
      if (!w->close()) {
        slotStatusMsg(i18n("Ready."));
        break;
      }
}

void KplApp::slotEditUndo()
{
  doc->restoreItems();
}

void KplApp::slotEditRedo()
{
  doc->restoreItems(false);
}

void KplApp::slotEditItems()
{
  if (!itemDlg) {
    itemDlg = new ItemDlg(doc);
    connect(itemDlg, SIGNAL(displayMessage(const QString &)),
            SLOT(slotStatusMsg(const QString &)));
  }
  itemDlg->show();
  itemDlg->raise();
}

void KplApp::slotViewZoomIn()
{
  view->fZoom = 0.01 * (QMIN(qRound(100 * view->fZoom) + 20, 2000));
  shrink();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotViewZoomOut()
{
  view->fZoom = 0.01 * (QMAX(qRound(100 * view->fZoom) - 20, 1));
  shrink();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotViewZoom()
{
  statusBar()->message(i18n("Setting magnification..."));
  ZoomDlg dlg(this);
  dlg.factor->setValue(qRound(100 * view->fZoom));
  if (dlg.exec() == dlg.Accepted) {
    view->fZoom = 0.01 * dlg.factor->interpretedValue();
    shrink();
  }
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotViewReload()
{
  statusBar()->message(i18n("Reloading file..."));
  autoplot->setChecked(true);
  doc->openDocument(doc->currentURL(), doc->fType);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotViewAutoReload()
{
  statusBar()->message(i18n("Reloading file periodically..."));
  RefreshDlg dlg(this);
  bool timerActive = timer->isActive();
  dlg.setWidgets(tRefresh, timerActive);
  if (dlg.exec() == dlg.Accepted) {
    int tRef;
    bool act;
    dlg.getValues(&tRef, &act);
    if (timerActive && ((tRef != tRefresh) || (act != timerActive)))
      timer->stop();
    if (act && ((tRef != tRefresh) || (act != timerActive))) {
      timer->start(1000 * tRef);
    }
    tRefresh = tRef;
  }
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotOptionsToolBar()
{
  toolBar("mainToolBar")->enable(KToolBar::Toggle);
}

void KplApp::slotOptionsStatusBar()
{
  if (statusBar()->isVisible())
	  statusBar()->hide();
  else
	  statusBar()->show();
}

void KplApp::slotOptionsAutoPlot()
{
  if (autoplot->isChecked())
    slotFilePlot();
}

void KplApp::slotOptionsShowSource()
{
  doc->aut.showSource = showSource->isChecked();
}

void KplApp::slotOptionsKeys()
{
  statusBar()->message(i18n("Key bindings..."));
  KKeyDialog::configureKeys(actionCollection(), xmlFile());
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotEditToolbars()
{
  statusBar()->message(i18n("Toolbar configuration..."));
  KEditToolbar dlg(actionCollection());
  if (dlg.exec())
    createGUI();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotOptionsBasic()
{
  statusBar()->message(i18n("Basic settings..."));
  BasicDlg dlg(this, &doc->aut);
  if (dlg.exec() == dlg.Accepted)
    shrink();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotOptionsAuto()
{
  statusBar()->message(i18n("Autoplot settings..."));
  AutoDlg dlg(this, doc);
  dlg.exec();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotOptionsAutoFit()
{
  statusBar()->message(i18n("Autofit settings..."));
  AutoFitDlg dlg(this, doc);
  dlg.exec();
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotOptionsSave()
{
  saveOptions(true);
  slotStatusMsg(i18n("Ready."));
}

void KplApp::slotStatusMsg(const QString &text)
{
  statusBar()->clear();
  statusBar()->changeItem(text, StatusMsg);
}

void KplApp::slotStatusHelpMsg(const QString &text)
{
  statusBar()->message(text, 3000);
}

void KplApp::slotOpenList(QStrList* list)
{
  for (const char* u = list->first(); u; u = list->next())
    if (u == list->getFirst())
      openDocumentURL(u);
    else {
      KplApp* kpl = new KplApp;
      if (kpl->doc->openDocument(u))
        kpl->show();
      else
        delete kpl;
    }
}

void KplApp::slotUpdate(bool updPlot, bool updList)
{
  updWin();
  if (updPlot)
    if (doc->itd.count())
      slotPlotCond();
    else
      view->eraseWidget();
  if (itemDlg && updList)
    itemDlg->updItemList();
}

void KplApp::updWin()
{
  if (doc->currentURL().fileName() == i18n("Untitled"))
    setCaption(doc->currentURL().fileName(), doc->isModified());
  else {
    recentFiles->addURL(doc->currentURL());
    if (doc->currentURL().isLocalFile())
      setCaption(doc->currentURL().path(), doc->isModified());
    else
      setCaption(doc->currentURL().url(), doc->isModified());
  }
  recentFiles->setEnabled(recentFiles->items().count());
  bool nItems = doc->itd.count();
  fileSaveAs->setEnabled(nItems);
  fileClose->setEnabled(nItems);
  filePlot->setEnabled(nItems);
  filePS->setEnabled(nItems);
  filePrint->setEnabled(nItems);
  fileSave->setEnabled(doc->isModified());
  editUndo->setEnabled(doc->undoAllowed());
  editRedo->setEnabled(doc->redoAllowed());
  viewReload->setEnabled((doc->currentURL().fileName() != i18n("Untitled")));
  plot = autoplot->isChecked() && nItems;
}

void KplApp::updStatus(int x, int y)
{
  if (statusBar()->isVisible()) {
    double xcm, ycm, xr, yr;
    if ((!view->i2cm(x, y, &xcm, &ycm)) && plot) {
      view->i2r(x, y, view->fxn, view->fyn, &xr, &yr);
      statusBar()->changeItem("  x = " + QString::number(xr, 'g', xPrec) +
                              ",  y = " += QString::number(yr, 'g', yPrec),
                              StatusVal);
    } else
      statusBar()->changeItem("", StatusVal);
    QCString s;
    statusBar()->changeItem(s.sprintf("%6.2f cm, %6.2f cm", xcm, ycm),
                            StatusPos);
  }
}

void KplApp::clearStatus()
{
  if (statusBar()->isVisible()) {
    statusBar()->changeItem("", StatusVal);
    statusBar()->changeItem("", StatusPos);
  }
}

void KplApp::shrink()
{
  view->layout();
  if ((QMAX(view->xPict, 120) < view->scrx) || (view->yPict < view->scry)) {
    int no = 1 + (toolBar()->sizeHint().width() - 1) / width();
    int w = QMIN(QMAX(view->xPict, 120), view->scrx) - view->scrx + width() + 2;
    int n = 1 + (toolBar()->sizeHint().width() - 1) / w;
    resize(w, QMIN(view->yPict, view->scry) - view->scry + height() +
           menuBar()->heightForWidth(w) - menuBar()->height() +
           (n - no) * (toolBar()->height() / no) + 2);
    newStatusBar();
  } else
    slotPlotCond();
}

void KplApp::showPopupMenu()
{
  ((QPopupMenu*) factory()->container("popup", this))->exec(QCursor::pos());
}
