/***************************************************************************
                          kpl.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPL_H
#define KPL_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <kmainwindow.h>
#include "kplnamespace.h"

class QStrList;
class KURL;
class KAccel;
class KAction;
class KListAction;
class KToggleAction;
class KRecentFilesAction;
class KplDoc;
class KplView;
class ItemDlg;

/**
  * Kpl application.
  * Provides the kpl top level main window with menubar, toolbar and statusbar.
  * @author Werner Stille
  */

class KplApp : public KMainWindow
{
  Q_OBJECT

public:
  /** Constructor. Creates a kpl window. */
  KplApp();
  /** Destructor. */
  ~KplApp();
	/**
	 * Returns a pointer to the current document connected to the KMainWindow
	 * instance and is used by the View class to access the document object's
	 * methods.
   * @return pointer to the current document.
   */	
	KplDoc* getDocument() const; 	

protected:
  /** Initializes menu and toolbar. */
  void initActions();
  /** Calculates precision for statusbar. */
  void calcStatusBar();
  /** Inserts statusbar items. */
  void initStatusBar();
  /** Creates and initialize new statusbar. */
  void newStatusBar();
	/**
   * initKeyAccel creates the keyboard accelerator items for the available slots and changes the menu accelerators.
   * @see KAccel
   */
	void initKeyAccel();
	/**
   * Initializes the document object of the main window that is connected to the view in initView().
   * @see initView();
   */
	void initDocument();
  /**
   * Creates the centerwidget of the KMainWindow instance and sets it as the view.
   */
  void initView();
  /**
   * Saves the options.
   * @param saveAll save everything when true.
   */
  void saveOptions(bool saveAll);
  /** Reads the options on init. */
  void readOptions();
  /**
   * Opens an existing URL or local file for reading.
   * @param type one of the following: UNKNOWN, DATAFILE, PLOTFILE.
   */
  void openRead(Kpl::FileTypes type);
  /**
   * Saves file using current or new name.
   * @param current true for save under current name.
   * @return true when file is saved.
   */
  bool save(bool current);
  /**
   * Warns at unsaved changes.
   * @return true when ready to close file.
   */
  virtual bool queryClose();
  /** Reimplemented for message box on last window exit. */
  virtual bool queryExit();
  /**
   * Saves the window properties for each open window during session end
   * to the session config file, including saving the currently
   * opened file by a temporary filename provided by KApplication.
   * @see KMainWindow#saveProperties
   */
  virtual void saveProperties(KConfig*);
  /**
   * Reads the session config file and restores the application's state
   * including the last opened files and documents by reading the
   * temporary files saved by saveProperties().
   * @see KMainWindow#readProperties
   */
  virtual void readProperties(KConfig*);
  KAction *fileSave, *fileSaveAs, *fileClose, *filePrint, *filePlot,
          *editUndo, *editRedo, *viewReload;
  KListAction* filePS;
  KToggleAction *showToolbar, *showStatusbar, *autoplot, *showSource,
                *boundingBox, *saveAbsPath, *warning, *saveAtEnd;
  KRecentFilesAction* recentFiles;
  KplView* view;
  KplDoc* doc;
  ItemDlg* itemDlg;
  KConfig* config;
  KAccel* key_accel;
  QTimer* timer;
  enum StatusId{StatusMsg = 1, StatusPos, StatusVal};
  bool plot;
  int mainHeight, mainWidth, tRefresh, xPrec, yPrec;
  double fz;

  friend KplView;

public slots:
	/**
	 * Opens a file.
   * @param url URL of the file.
	 */
  void openDocumentURL(const KURL& url);

protected slots:
  /** Creates a new plot in the current view. */
  void slotFileNew();
  /** Opens an existing plot file. */
  void slotFileOpenPlot();
  /** Opens an existing data file. */
  void slotFileOpenData();
  /** Saves the current plot file. */
  void slotFileSave();
  /** Saves the current plot as a file with different name. */
  void slotFileSaveAs();
  /** Closes the current file. */
  void slotFileClose();
  /** Plots the current file. */
  void slotFilePlot();
  /** Plots the current file conditionally. */
  void slotPlotCond();
  /**
   * Generates PostScript file.
   * @param landscape != 0 for landscape orientation.
   */
  void slotFilePS(int landscape);
  /** Prints the current file. */
  void slotFilePrint();
  /** Opens a new kpl window. */
  void slotFileNewWindow();
  /** Closes the current window. */
  void slotFileCloseWindow();
  /** Exits the application. */
  void slotFileQuit();
  /** Restores previous item from backup. */
  void slotEditUndo();
  /** Restores next item from backup. */
  void slotEditRedo();
  /** Edits plot items. */
  void slotEditItems();
  /** Increases magnification. */
  void slotViewZoomIn();
  /** Decreases magnification. */
  void slotViewZoomOut();
  /** Sets magnification. */
  void slotViewZoom();
  /** Reloads file. */
  void slotViewReload();
  /** Reloads file periodically. */
  void slotViewAutoReload();
  /** Toggles the visibility of the toolbar. */
  void slotOptionsToolBar();
  /** Toggles the visibility of the statusbar. */
  void slotOptionsStatusBar();
  /** Toggles the autoplot option. */
  void slotOptionsAutoPlot();
  /** Toggles the show function source option. */
  void slotOptionsShowSource();
  /** Displays the key bindings dialog. */
  void slotOptionsKeys();
  /** Displays the toolbars configure dialog. */
  void slotEditToolbars();
  /** Displays the basic settings dialog. */
  void slotOptionsBasic();
  /** Displays the autoplot settings dialog. */
  void slotOptionsAuto();
  /** Displays the autofit settings dialog. */
  void slotOptionsAutoFit();
  /** Saves options. */
  void slotOptionsSave();
  /**
   * Changes the status message.
   * @param text pointer to new message text.
   */
  void slotStatusMsg(const QString& text);
  /**
   * Changes the status message of the whole statusbar temporarily.
   * @param text pointer to temporary message text.
   */
  void slotStatusHelpMsg(const QString& text);
  /**
   * Opens files from URL list.
   * @param list pointer to list with URLs.
   */
  void slotOpenList(QStrList* list);
  /** Updates titlebar, menus, plot, and item list.
   * @param updPlot true when plot should be uddated.
   * @param updList true when item list should be uddated.
   */
  void slotUpdate(bool updPlot, bool updList);
  /** Updates titlebar and menus. */
  void updWin();
  /**
   * Updates mouse position on statusbar.
   * @param x horizontal position of the mouse cursor.
   * @param y vertical position of the mouse cursor.
   */
  void updStatus(int x, int y);
  /** Clears mouse position information on statusbar. */
  void clearStatus();
  /** Shrinks window. */
  void shrink();
  /** Shows main popup menu. */
  void showPopupMenu();
};

#endif // KPL_H
