/***************************************************************************
                          kpldoc.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <math.h>
#include <qdir.h>
#include <klocale.h>
#include <ksimpleconfig.h>
#include <kmessagebox.h>
#include <kio/job.h>
#include <kio/netaccess.h>
#include "kpldoc.h"
#include "fitdlg.h"
#include "frameitem.h"
#include "arrayitem.h"
#include "funitem.h"
#include "legenditem.h"
#include "utils.h"

KplDoc::KplDoc(QObject* parent) :
 QObject(parent)
{
  fType = Kpl::Unknown;
  itd.setAutoDelete(true);
  itb.setAutoDelete(true);
}

KplDoc::~KplDoc()
{
  freeAllItems();
  itb.clear();
}

void KplDoc::newDocument()
{
  docURL.setFileName(i18n("Untitled"));
  freeAllItems();
  aut.nyLeg = 0;
}

bool KplDoc::saveDocument(const KURL& url, bool abs)
{
  FILE* f;
  if (url.isLocalFile())
    f = fopen(url.path(), "w");
  else
    f = fopen(QDir::homeDirPath() + "/kpltmp.tmp", "w");
  if (f) {
    fprintf(f, "[Items]\nnitems=%i\n", itd.count());
    for (unsigned int i = 0; i < itd.count(); i++) {
      fprintf(f, "\n[Item %i]\n", i);
      itd.at(i)->writePlo(f, url, abs);
    }
    fclose(f);
    if (!url.isLocalFile()) {
      job = KIO::file_copy(KURL(QDir::homeDirPath() + "/kpltmp.tmp"), url, -1,
                           true);
      connectIOfinished();
    }
    setModified(false);
    fType = Kpl::Plot;
  } else
    KMessageBox::sorry((QWidget *) parent(),
                       i18n("Writing to this file not possible."));
  return (f);
}

void KplDoc::setModified(bool m)
{
  modified = m;
}

bool KplDoc::isModified() const
{
  return modified;
}

void KplDoc::setCurrentDir(const KURL& url)
{
  QFileInfo fi(url.path());
  KURL u = url;
  u.setPath(fi.dirPath(true));
  m_dir = u.url(+1);
}

const QString& KplDoc::currentDir() const
{
  return m_dir;
}

void KplDoc::setCurrentURL(const KURL& url)
{
  docURL = url;
  setCurrentDir(url);
}

const KURL& KplDoc::currentURL() const
{
  return docURL;
}

void KplDoc::freeAllItems()
{
  itd.clear();
  setModified(false);
}

void KplDoc::newItem(Kpl::ItemTypes ityp)
{
  switch (ityp) {
    case Kpl::Frame:
      itd.append(new FrameItem(&aut));
      break;
    case Kpl::Array:
      itd.append(new ArrayItem(&aut));
      break;
    case Kpl::Function:
      itd.append(new FunItem(&aut));
      break;
    case Kpl::ParFunction:
      itd.append(new ParFunItem(&aut));
      break;
    case Kpl::Legend:
      itd.append(new LegendItem(&aut));
  }
}

bool KplDoc::autoScale(FrameItem* fd, ArrayItem* ad)
{
  if ((ad->x) && (aut.ixAuto < ad->ncols) && (aut.iyAuto < ad->ncols)) {
    if (Utils::autoFile(&fd->xmi, &fd->xma, &fd->xtic, &fd->mticx, &fd->ndigx,
                        &fd->iex, &ad->fx, ad->x[aut.ixAuto], ad->nrows,
                        aut.autoNorm, 0.2))
      return true;
    if (Utils::autoFile(&fd->ymi, &fd->yma, &fd->ytic, &fd->mticy, &fd->ndigy,
                        &fd->iey, &ad->fy, ad->x[aut.iyAuto], ad->nrows,
                        aut.autoNorm, 0.2 * aut.xlAuto / aut.ylAuto))
      return true;
    if (aut.autohPath)
      strncpy(fd->sh, ad->path, sizeof(aut.shAuto) - 1);
    int minCol = ad->ncols - 1;
    ad->ix = QMIN(aut.ixAuto, minCol);
    ad->iy = QMIN(aut.iyAuto, minCol);
    ad->ie = QMIN(aut.ieAuto, minCol);
    ad->symb = aut.autoSymb;
    ad->istart = 0;
    ad->n = ad->nrows;
    ad->errbars = aut.autoErr & (ad->ie < ad->ncols);
    return false;
  }
  return true;
}

void KplDoc::autoFit(ArrayItem* ad, FunItem* fd)
{
  strcpy(fd->pathy, aut.fitPath);
  strcpy(fd->namey, aut.fitName);
  memcpy(fd->py, aut.pFit, KPL_NPMAX * sizeof(double));
  fd->symb = aut.fitSymb;
  fd->color = aut.fitColor;
  fd->tmin = ad->x[ad->ix][ad->istart];
  fd->tmax = ad->x[ad->ix][ad->istart + ad->n - 1];
  fd->fx = ad->fx;
  fd->fy = ad->fy;
  if (Utils::getFuncAddr(fd->pathy, fd->namey, &fd->hmody, &fd->fkty))
    return;
  int mode = 0;
  if (aut.fitShowDlg)
    mode |= FitDlg::ShowDlg;
  if (aut.fitSavePar)
    mode |= FitDlg::SavePar;
  if (aut.fitFollow)
    mode |= FitDlg::Follow;
  QList<ArrayItem> arr;
  arr.append(ad);
  QList<FunItem> fun;
  fun.append(fd);
  FitDlg dlg((QWidget*) parent(), this, &arr, &fun, mode);
  dlg.slotFit();
  if (aut.fitShowDlg) {
    backupItems();
    dlg.exec();
  } else
    dlg.slotApply();
  return;
}

bool KplDoc::openDocument(const KURL& url, Kpl::FileTypes type)
{
  fType = type;
  if (fType == Kpl::Unknown) {
    QString s = url.fileName();
    fType = ((s.length() - s.findRev(".plo", -1, false)) == 4) ?
            Kpl::Plot : Kpl::Data;
  }
  bool success;
  if (fType == Kpl::Data) {
    freeAllItems();
    newItem(Kpl::Frame);
    newItem(Kpl::Array);
    if (aut.autoFit)
      newItem(Kpl::Function);
    if (url.isLocalFile())
      strcpy(((ArrayItem *) itd.at(1))->path, url.path());
    else
      strcpy(((ArrayItem *) itd.at(1))->path, url.url());
    int nrows = ((ArrayItem *) itd.at(1))->readFile();
    success = nrows;
    if (success) {
      if (autoScale((FrameItem *) itd.at(0), (ArrayItem *) itd.at(1)))
        success = false;
      else {
        setCurrentURL(url);
        if (aut.autoFit)
          autoFit((ArrayItem *) itd.at(1), (FunItem *) itd.at(2));
        setModified();
      }
    }
  } else {
    QString tmpFile;
    success = KIO::NetAccess::download(url, tmpFile);
    if (success) {
      QFileInfo f2(tmpFile);
      success = f2.isReadable();
      if (success) {
        newDocument();
        KSimpleConfig plo(tmpFile, true);
        plo.setGroup("Items");
        if (int nItems = QMAX(plo.readNumEntry("nitems"), 0)) {
          const char *itemName[] = {"", "FRAMEITEM", "ARRAYITEM", "FUNITEM",
                                    "PARFUNITEM", "LEGENDITEM"};
          setCurrentURL(url);
          for (int i = 0; i < nItems; i++) {
            QString s;
            s.sprintf("Item %i", i);
            plo.setGroup(s);
            s = plo.readEntry("Type");
            int iType;
            for (iType = Kpl::Frame; iType <= Kpl::Legend; iType++)
              if (s == itemName[iType])
                break;
            switch (iType) {
              case Kpl::Frame:
                itd.append(new FrameItem(&plo, &aut));
                break;
              case Kpl::Array:
                itd.append(new ArrayItem(&plo, &aut, url));
                break;
              case Kpl::Function:
                itd.append(new FunItem(&plo, &aut, url));
                break;
              case Kpl::ParFunction:
                itd.append(new ParFunItem(&plo, &aut, url));
                break;
              case Kpl::Legend:
                itd.append(new LegendItem(&plo, &aut));
            }
          }
        } else
          success = false;
      }
      KIO::NetAccess::removeTempFile(tmpFile);
    }
  }
  if (success)
    backupItems();
  else {
    restoreItems();
    KMessageBox::error((QWidget *) parent(),
                       QString(i18n("while loading the file\n")) + url.path());
  }
  return success;
}

void KplDoc::connectIOfinished()
{
  connect(job, SIGNAL(result(KIO::Job *)), this, SLOT(slotIOfinished()));
}

bool KplDoc::restoreItems(bool undo)
{
  QList<KplItem>* itc = 0;
  if (undo) {
    if (undoAllowed())
      itc = itb.prev();
  } else {
    if (redoAllowed())
      itc = itb.next();
  }
  if (itc) {
    itd.clear();
    for (unsigned int i = 0; i < itc->count(); i++)
    itd.append(itc->at(i)->copy());
    emit modelChanged(true, true);
    return true;
  } else
    return false;
}

bool KplDoc::undoAllowed() const
{
  if (!itb.current())
    return false;
  return (itb.current() != itb.getFirst());
}

bool KplDoc::redoAllowed() const
{
  if (!itb.current())
    return false;
  return (itb.current() != itb.getLast());
}

void KplDoc::slotIOfinished()
{
  if (job->error()) {
    job->showErrorDialog();
    QFileInfo fi(job->destURL().path());
    if (fi.extension().lower() == "plo") {
      setModified();
      emit modelChanged(false, false);
    }
  }
  remove(QDir::homeDirPath() + "/kpltmp.tmp");
}

void KplDoc::backupItems(bool list)
{
  QList<KplItem>* itc = itb.current();
  if (itc) {
    if (itc == itb.getLast()) {
      if (itb.count() >= 20)
        itb.removeFirst();
    } else
      do
        itb.removeLast();
      while (itc != itb.current());
  }
  itb.append(new QList<KplItem>);
  for (unsigned int i = 0; i < itd.count(); i++)
    itb.current()->append(itd.at(i)->copy());
  emit modelChanged(true, list);
}
