/***************************************************************************
                          utils.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef UTILS_H
#define UTILS_H

#include <stdio.h>
#include <qdialog.h>
#include <kurl.h>
#include "kplnamespace.h"

class QLineEdit;
class QWidget;
class QString;
class QListBox;
class KplDoc;

/**
  * Choose function dialog class. Provides a dialog for the selection
  * of a function from list of functions in a shared library.
  * @author Werner Stille
  */
class ChooseFuncDlg : public QDialog {
  Q_OBJECT

public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param f pointer to stream opened for input from a file containing
   * a list of functions obtained using nm.
   * @param func pointer to QLineEdit widget for the function name.
   */
  ChooseFuncDlg(QWidget* parent, FILE* f, QLineEdit* func);
  /** Destructor. */
  ~ChooseFuncDlg();

protected:
  QLineEdit* func;
  QListBox* funcList;

protected slots:
  /**
   * Accepts selection and leaves the dialog.
   * @param i index of selected function
   */
  void slotList(int i);
  /** Accepts selection and leaves the dialog. */
  void slotOK();
};

/**
  * Edit parameter dialog class. Provides a dialog for the editing
  * the parameters of a function.
  * @author Werner Stille
  */
class EditParameterDlg : public QDialog {
  Q_OBJECT

public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param m pointer to KplDoc model object.
   * @param p parameter array.
   * @param file absolute path of the function library.
   * @param func function name.
   */
  EditParameterDlg(QWidget* parent, KplDoc* m, double* p, const char* file,
                   const char* func);
  /** Destructor. */
  ~EditParameterDlg();

protected:
  double* p;
  QLineEdit* par[KPL_NPMAX];

protected slots:
  /** Accepts changes and leaves the dialog. */
  void slotOK();
};

/**
  * Utility class. Provides static utility functions in a namespace
  * given by the class.
  * @author Werner Stille
  */
class Utils {
public:
  /**
   * Retrieves URL for saving.
   * @param parent parent widget.
   * @param url KURL receiving the chosen file URL.
   * @param filter file name filter for file dialog.
   * @param m KplDoc model object.
   * @return true when file can be saved.
   */
  static bool getWriteURL(QWidget* parent, KURL& url, const char* filter,
                          KplDoc* m);
  /**
   * Opens a file selection dialog.
   * @param e QLineEdit widget to display the filename.
   * @param m KplDoc model object.
   */
  static void getFile(QLineEdit* e, KplDoc* m);
  /**
   * Opens a function selection dialog.
   * @param parent parent widget.
   * @param name QLineEdit widget displaying the library filename.
   * @param func QLineEdit widget to display the function name.
   */
  static void getFunc(QWidget* parent, QLineEdit* name, QLineEdit* func);
  /**
   * Opens a dialog to select a function parameter file.
   * @param parent parent widget.
   * @param p double array for parameters.
   * @param m KplDoc model object.
   */
  static void getPar(QWidget* parent, double* p, KplDoc* m);
  /**
   * Opens a dialog to save function parameters in a file.
   * @param parent parent widget.
   * @param p double array with parameters.
   * @param m KplDoc model object.
   */
  static void savePar(QWidget* parent, double* p, KplDoc* m);
  /**
   * Gets module handle and address of a shared library function.
   * @param path path of the library.
   * @param parent parent widget.
   * @param name function name.
   * @param hmod pointer to module handle.
   * @param fkt pointer to function address.
   */
  static bool getFuncAddr(const QString& path, const QString& name,
                          void** hmod, double (**fkt)(double, const double*));
  /**
   * Calculates minimum and maximum values of an array.
   * @param xmin pointer to minimum value.
   * @param xmax pointer to maximum value.
   * @param x double array.
   * @param n size of array.
   * @return true for error.
   */
  static bool minMaxFile(double* xmin, double* xmax, const double* x, int n);
  /**
   * Calculates normalization factor.
   * @param a absolute double value.
   * @param ie pointer to int for logarithm of normalization factor.
   * @param f pointer to double for normalization factor.
   */
  static void expo(double a, int* ie, double* f);
  /**
   * Calculates normalization factor for an array.
   * @param ie pointer to int for logarithm of normalization factor.
   * @param f pointer to double for normalization factor.
   * @param x double array.
   * @param n size of array.
   */
  static void expoFile(int* ie, double* f, const double* x, int n);
  /**
   * Calculates normalization factor for a function.
   * @param fkt function.
   * @param ie pointer to int for logarithm of normalization factor.
   * @param f pointer to double for normalization factor.
   * @param xmin minimum argument value.
   * @param xmax maximum argument value.
   * @param dx argument stepsize.
   * @param p double array with parameters.
   */
  static void expoFunc(double (*fkt)(double, const double*), int* ie,
                       double* f, double xmin, double xmax, double dx,
                       const double* p);
  /**
   * Calculates scaling parameters.
   * @param xmi pointer to double for rounded minimum value.
   * @param xma pointer to double for rounded maximum value.
   * @param tic pointer to double for major tic distance.
   * @param mtic pointer to double for intervals between major tics.
   * @param ndig pointer to double for number of places after decimal point
   * for lettering.
   * @param xmin minimum value.
   * @param xmax maximum value.
   * @param fx normalization factor.
   * @param c scaling constant, desired value for tic / (xma - xmi).
   * @return true when xmin and xmax are equal.
   */
  static bool autoSc(double* xmi, double* xma, double* tic, int* mtic,
                     int* ndig, double xmin, double xmax, double fx, double c);
  /**
   * Calculates scaling parameters for an array.
   * @param xmi pointer to double for rounded minimum value.
   * @param xma pointer to double for rounded maximum value.
   * @param tic pointer to double for major tic distance.
   * @param mtic pointer to double for intervals between major tics.
   * @param ndig pointer to double for number of places after decimal point
   * for lettering.
   * @param ie pointer to int for logarithm of normalization factor.
   * @param fn pointer to double for normalization factor.
   * @param x double array.
   * @param n true when normalization is wanted.
   * @param norm size of array.
   * @param c scaling constant, desired value for tic / (xma - xmi).
   * @return true when values are equal.
   */
  static bool autoFile(double* xmi, double* xma, double* tic, int* mtic,
                       int* ndig, int* ie, double* fn, const double* x, int n,
                       bool norm, double c);
  /**
   * Calculates scaling parameters for a function.
   * @param xmi pointer to double for rounded minimum value.
   * @param xma pointer to double for rounded maximum value.
   * @param tic pointer to double for major tic distance.
   * @param mtic pointer to double for intervals between major tics.
   * @param ndig pointer to double for number of places after decimal point
   * for lettering.
   * @param fkt function.
   * @param xmin minimum argument value.
   * @param xmax maximum argument value.
   * @param dx argument stepsize.
   * @param p double array with parameters.
   * @param fn normalization factor.
   * @param c scaling constant, desired value for tic / (xma - xmi).
   * @return true when values are equal.
   */
  static bool autoFunc(double* ymi, double* yma, double* tic, int* mtic,
                       int* ndig, double (*fkt)(double, const double*),
                       double xmin, double xmax, double dx, const double *p,
                       double fn, double c);
  /**
   * Converts double to integer.
   * @param a double value.
   * @return integer result.
   */
  static int int9(double a);
  /**
   * Converts Qt1 QRgb to QRgb.
   * @param u Qt1 rgb value.
   * @return QRgb value for actual Qt version.
   */
  static unsigned rgbQt(unsigned u);
  /**
   * Converts QRgb to Qt1 QRgb.
   * @param rgb QRgb value of actual Qt version.
   * @return QRgb value for actual Qt1 version.
   */
  static unsigned rgbQt1(unsigned rgb);
  /**
   * Constructs absolute file path.
   * @param url URL of plot file.
   * @param path file path.
   */
  static void absPath(const KURL& url, char* path);
  /**
   * Constructs relative file path (if possible).
   * @param url URL of plot file.
   * @param rel pointer to string for relative path.
   * @param path file path.
   * @param abs return absolute path if true.
   * @return relative path (if possible).
   */
  static const char* relPath(const KURL& url, char* rel, const char* path,
                             bool abs);
};

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Finds the inverse of a n by n matrix.
 * @param A input matrix.
 * @param X output matrix.
 * @param n numbers of rows and columns.
 * @param B scratch pad array of length n.
 * @param IPS scratch pad arra of length n.
 * @return 0 for success.
 */
int minv(double A[], double X[], int n, double B[], int IPS[]);
/**
 * Calculates complemented incomplete gamma integral.
 * @param a argument of gamma function.
 * @param x upper boundary of integral.
 * @return complemented incomplete gamma integral.
 */
double igamc(double a, double x);
/**
 * Calculates incomplete gamma integral.
 * @param a argument of gamma function.
 * @param x upper boundary of integral.
 * @return incomplete gamma integral.
 */
double igam(double a, double x);
/**
 * Calculates logarithm of gamma function.
 * @param x argument of gamma function.
 * @return logarithm of gamma function.
 */
double lgam(double x);

#ifdef __cplusplus
}

#endif

#endif
